#!/usr/bin/env python
# encoding: utf-8


import pandas as pd
import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader


class PredictDataset:
    def __init__(self, df, config):
        self.df = df
        self.lookback = config.get("lookback")
        self.target = config.get("target")
        self.external_features = config.get("external_features")
        self.sample_size = len(self.df) - self.lookback

    def __len__(self):
        return self.sample_size

    def pop(self, i, preds):
        kk = (
            self.df[[self.target] + self.external_features]
            .copy()
            .values[i : i + self.lookback, :]
        )
        if preds != []:
            predicted = min(self.lookback, len(preds))
            # kk[-len(preds):, 0] = [3*i for i in range(len(preds))]
            kk[-predicted:, 0] = preds[-self.lookback :]
        kk = kk.astype("float32")
        return kk


class TimeSeriesDataset(Dataset):
    def __init__(self, dataset, config):
        self.df = dataset
        self.external_features = config.get("external_features")
        self.target = config.get("target")
        self.lookback = config.get("lookback")
        self.forecast_horizon = config.get("forecast_horizon")
        self.sample_size = len(self.df) - self.lookback - self.forecast_horizon

    def __len__(self):
        return self.sample_size

    def build(self):
        self.x = []
        self.y = []
        for i in range(0, self.sample_size):
            sample_x = (
                self.df[[self.target] + self.external_features]
                .copy()
                .values[i : i + self.lookback, :]
            )
            sample_y = (
                self.df[self.target]
                .copy()
                .values[
                    i
                    + self.lookback : i
                    + self.lookback
                    + self.forecast_horizon
                ]
            )
            # uniform
            # offset = sample_x[0, 0]
            # sample_x[:, 0] -= offset
            # sample_y[0] -= offset
            self.x.append(sample_x)
            self.y.append(sample_y)
        # self.x = np.asarray(self.x)
        # self.y = np.asarray(self.y)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # print(type(self.x))
        temp_x = self.x[idx].astype("float32")
        temp_y = self.y[idx].astype("float32")

        # sample = (np.array()self.x[idx], self.y[idx])
        sample = (temp_x, temp_y)
        # print(sample[0].dtype)

        return sample


def make_timeline(df, freq="D", periods=100):
    if ("ds" not in df) or ("y" not in df):
        raise ValueError("xxx")
    history = pd.to_datetime(df["ds"]).sort_values()
    last_date = history.max()

    extend_dates = pd.date_range(
        start=last_date, periods=periods + 1, freq=freq, closed="right"
    )
    extend_df = pd.DataFrame({"ds": extend_dates})
    full_dates = pd.concat([df, extend_df]).reset_index(drop=True)

    return full_dates

    # return pd.DataFrame({"ds": full_dates})


#    df["y"] = pd.to_numeric(df["y"])
#    if np.isinf(df["y"].values).any():
#        raise ValueError("")
#    if df["ds"].dtype == np.int64:
#        df["ds"] = df["ds"].astype(str)
#    df["ds"] = pd.to_datetime(df["ds"])
#    if df["ds"].isnull().any():
#        raise ValueError("xxx")



#!/usr/bin/env python
# encoding: utf-8

import pytorch_lightning as pl


class BaseTrainer(pl.Trainer):
    def __init__(self, *args, **kargs):
        super(BaseTrainer, self).__init__(*args, **kargs)



#!/usr/bin/env python
# encoding: utf-8

from argparse import ArgumentParser

from sibyldeep.models import BaseModel

from pytorch_lightning import Trainer


import pandas as pd

batch_size = 1
forecast_horizon = 1
lookback = 28
n_epochs = 5

train_true_y = []
train_pred_y = []

df = pd.read_csv("./data.csv")
df.columns = ["ds", "y"]
df["ext"] = 1
df = df[:50]


def train(args):
    sibyl_model = BaseModel(args)
    trainer = Trainer(args)
    trainer.fit(sibyl_model)


if __name__ == "__main__":
    hyper_params_parser = ArgumentParser(description="parser")
    # model level
    hyper_params_parser = BaseModel.add_model_specific_args(hyper_params_parser)

    # trainer level
    hyper_params_parser = Trainer.from_argparse_args(hyper_params_parser)

    hyper_params = hyper_params_parser.parse_args()
    train(hyper_params)



#!/usr/bin/env python
# encoding: utf-8

from mssibyl.models import BaseModel
from mssibyl.trainer import BaseTrainer
from mssibyl.datasets.utils import make_timeline

import pandas as pd

df = pd.read_csv("./data.csv")
df.columns = ["ds", "y"]
df = df[:1000]

df_with_future = make_timeline(df, freq="H", periods=300)

args = {
    "lookback": 48,
    "forecast_horizon": 1,
    "data": df_with_future,
    "target": "y",
}

sibyl_model = BaseModel(args)
trainer = BaseTrainer(
    max_epochs=100, gpus=1, show_progress_bar=False, early_stop_callback=True
)
trainer.fit(sibyl_model)
preds = sibyl_model.predict(interval=True, include_history=True)

preds.to_csv("res.csv", index=False)



#!/usr/bin/env python
# encoding: utf-8

import pandas as pd
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import pytorch_lightning as pl
from mssibyl.datasets.utils import TimeSeriesDataset, PredictDataset
from argparse import ArgumentParser
from sklearn.preprocessing import MinMaxScaler


class BaseModel(pl.LightningModule):
    def __init__(self, model_config):
        super(BaseModel, self).__init__()
        self.hparams = self._build_model_specific_hparams(model_config)
        self.sc = MinMaxScaler(feature_range=(0, 0.5))

        self.target = self.hparams.get("target")
        self.forecast_horizon = self.hparams.get("forecast_horizon")
        self.lookback = self.hparams.get("lookback")
        self.num_features = len(self.hparams.get("external_features")) + 1
        self.hidden_size = self.hparams.get("hidden_size")
        self.bidirectional = False

        if isinstance(self.hparams.get("data"), str):
            self.raw_df = pd.read_csv(self.hparams.get("data"))
        else:
            self.raw_df = self.hparams.get("data")

        self.full_df = self.raw_df.copy()

        self.full_df[self.target] = self.sc.fit_transform(
            self.full_df[[self.target]]
        )

        self.encoder = nn.LSTM(
            input_size=self.num_features,
            hidden_size=self.hidden_size,
            num_layers=1,
            dropout=0.0,
            bidirectional=self.bidirectional,
            batch_first=True,
        )
        self.decoder = nn.LSTM(
            input_size=self.hidden_size,
            hidden_size=self.hidden_size // 4,
            num_layers=1,
            dropout=0.0,
            bidirectional=self.bidirectional,
            batch_first=True,
        )
        self.predictor = nn.Linear(
            in_features=self.hidden_size // 4,
            out_features=self.forecast_horizon,
        )
        self.loss_fn = nn.MSELoss()

    def forward(self, x, batch_size=1):

        output, _ = self.encoder(x)

        output = F.dropout(output, p=0.5, training=True)

        output, state = self.decoder(output)
        output = F.dropout(output, p=0.5, training=True)
        output = self.predictor(state[0].squeeze(0))

        return output

    def prepare_data(self):
        self.history = self.full_df.loc[
            self.full_df[self.target].notnull()
        ].copy()

        prepare_for_future = self.history[-self.lookback :]
        self.future = self.full_df.loc[
            self.full_df[self.target].isnull()
        ].copy()
        full_future = pd.concat([prepare_for_future, self.future]).reset_index(
            drop=True
        )

        split_ratio = 0.2

        threshold = int(len(self.history) * (1 - split_ratio))
        df_train = self.history[:threshold]
        df_valid = self.history[threshold:]
        self.dataset_train = TimeSeriesDataset(df_train, self.hparams)
        self.dataset_train.build()
        self.dataset_val = TimeSeriesDataset(df_valid, self.hparams)
        self.dataset_val.build()
        self.dataset_predict = PredictDataset(full_future, self.hparams)
        # self.dataset_predict = TimeSeriesDataset(full_future, self.hparams)
        # self.dataset_predict.build()

    def train_dataloader(self):
        return DataLoader(self.dataset_train, batch_size=1, num_workers=8)

    def val_dataloader(self):
        return DataLoader(self.dataset_val, batch_size=1, num_workers=8)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer

    def training_step(self, train_batch, batch_idx):
        x, y = train_batch
        logits = self.forward(x)
        loss = F.mse_loss(logits, y)
        logs = {"train_loss": loss}
        return {"loss": loss, "log": logs}

    def validation_step(self, valid_batch, batch_idx):
        x, y = valid_batch
        logits = self.forward(x)
        loss = F.mse_loss(logits, y)
        return {"val_loss": loss}

    def validation_epoch_end(self, outputs):
        avg_loss = torch.stack([x["val_loss"] for x in outputs]).mean()
        tensorboard_logs = {"val_loss": avg_loss}
        # return {"val_loss": avg_loss, "log": tensorboard_logs}
        return {"val_loss": avg_loss}

    def predict(
        self,
        inference_device=None,
        interval=True,
        num_repeats=5,
        include_history=True,
    ):
        if inference_device is None:
            if next(self.parameters()).is_cuda is True:
                inference_device = "cuda:0"
            else:
                inference_device = "cpu"

        repeated_preds = []
        for e in range(num_repeats):
            preds = []
            for i in range(len(self.dataset_predict)):
                x = self.dataset_predict.pop(i, preds)
                x = x.reshape(-1, self.lookback, self.num_features)
                xx = torch.Tensor(x).to(inference_device)
                pred = self.forward(xx)
                preds.append(pred.item())
            repeated_preds.append(preds)
        avg_preds = np.mean(np.array(repeated_preds), axis=0).reshape(-1)

        if interval is True:
            std_preds = np.std(np.array(repeated_preds), axis=0).reshape(-1)
            lower_bound = avg_preds - 2 * std_preds
            upper_bound = avg_preds + 2 * std_preds

            self.future["upper_bound"] = upper_bound
            self.future["lower_bound"] = lower_bound
            self.future["upper_bound"] = self.sc.inverse_transform(
                self.future[["upper_bound"]]
            )
            self.future["lower_bound"] = self.sc.inverse_transform(
                self.future[["lower_bound"]]
            )

        self.future[self.target] = avg_preds
        self.future[self.target] = self.sc.inverse_transform(
            self.future[[self.target]]
        )
        if include_history is True:
            full_predicted = pd.concat([self.raw_df, self.future])
        else:
            full_predicted = self.future

        return full_predicted

    def _build_model_specific_hparams(self, extra):
        default_hparas = {
            "hidden_size": 256,
            "forecast_horizon": 1,
            "lookback": 5,
            "external_features": [],
        }

        for k, v in extra.items():
            default_hparas[k] = v

        return default_hparas



#!/usr/bin/env python
# encoding: utf-8

from types import SimpleNamespace
from argparse import ArgumentParser



def build_args(args_dict):
    parser = ArgumentParser()
    for k,v in args_dict.items():
        print(k, v)
        parser.add_argument("--{}".format(k), type=type(v))

    parser.add_argument("--aa", type=int, default=1)

    return parser




# Copyright (c) Microsoft Corporation
# This source code is licensed under the MIT license

from os import path
from setuptools import setup, find_packages

with open(
    path.join(path.abspath(path.dirname(__file__)), "README.md"),
    encoding="utf-8",
) as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines()

setup(
    name="mssibyl",
    version="0.1.1.dev0",
    license="MIT",
    description="Time Series Forecasting Tool for Machine Learning Researchers with Nerual Process",
    packages=find_packages(),
    author="More Zhou",
    author_email="jinyzho@microsoft.com",
    keywords=["time series forecasting", "deep learning", "pytorch"],
    install_requires=["numpy", "pytorch-lightning", "pandas"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)
